<?php

namespace Drupal\helper;

use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Cache\UseCacheBackendTrait;
use DrupalFinder\DrupalFinder;

/**
 * Helper for working with Drupal's composer file.
 */
class Composer {

  use UseCacheBackendTrait;

  /**
   * The DrupalFinder.
   *
   * @var \DrupalFinder\DrupalFinder
   */
  public $drupalFinder;

  /**
   * Constructs a Composer helper.
   */
  public function __construct(CacheBackendInterface $cache_backend = NULL) {
    // Having issues injecting this via services, so optional parameter.
    $this->cacheBackend = $cache_backend ?? \Drupal::service('cache.helper_composer');
    $this->drupalFinder = new DrupalFinder();
    if (!$this->drupalFinder->locateRoot(DRUPAL_ROOT)) {
      throw new \RuntimeException("Unable to find Composer root directory from " . DRUPAL_ROOT);
    }
  }

  /**
   * Get the actual filename for composer.json.
   *
   * @return string
   *   The composer.json filename.
   */
  public static function getComposerFilename(): string {
    return trim(getenv('COMPOSER')) ?: 'composer.json';
  }

  /**
   * Get the actual filename for composer.lock.
   *
   * @return string
   *   The composer.lock filename.
   */
  public static function getComposerLockFilename(): string {
    $composer_filename = static::getComposerFilename();
    return pathinfo($composer_filename, PATHINFO_FILENAME) . '.lock';
  }

  /**
   * Gets contents of a composer JSON file.
   *
   * @param string|null $file
   *   The file to read, defaults to composer.json.
   * @param string|null $directory
   *   The directory to use as the working directory. If not provided this will
   *   attempt to use DrupalFinder to find the composer root directory.
   *
   * @return array
   *   The decoded JSON contents.
   *
   * @throws \RuntimeException
   * @throws \JsonException
   */
  public function getJson(?string $file = NULL, ?string $directory = NULL): array {
    $file ??= static::getComposerFilename();
    $directory ??= $this->drupalFinder->getComposerRoot();

    if (str_starts_with($directory, 'vendor/')) {
      $directory = str_replace('vendor/', $this->drupalFinder->getVendorDir(), $directory);
    }

    $uri = $directory . '/' . $file;
    if (!is_file($uri) || !is_readable($uri)) {
      throw new \RuntimeException("The file {$uri} does not exist or is not readable.");
    }

    if ($cache = $this->cacheGet($uri)) {
      // Perform an additional check on filemtime to ensure results are current.
      if (filemtime($uri) < $cache->created) {
        return $cache->data;
      }
    }

    $contents = file_get_contents($uri);
    if ($contents === FALSE) {
      throw new \RuntimeException("The file {$uri} could not be read.");
    }
    $contents = json_decode($contents, TRUE, 512, JSON_THROW_ON_ERROR);
    $this->cacheSet($uri, $contents);
    return $contents;
  }

  /**
   * Writes contents to a composer JSON file.
   *
   * @param mixed $data
   *   The contents to write to a file. If not a string, this will JSON encode
   *   the data before writing to the file.
   * @param string|null $file
   *   The file to read, defaults to composer.json.
   * @param string|null $directory
   *   The directory to use as the working directory. If not provided this will
   *   attempt to use DrupalFinder to find the composer root directory.
   *
   * @throws \RuntimeException
   * @throws \JsonException
   */
  public function writeJson(mixed $data, ?string $file = NULL, ?string $directory = NULL): void {
    $file ??= static::getComposerFilename();
    $directory ??= $this->drupalFinder->getComposerRoot();

    if (str_starts_with($directory, 'vendor/')) {
      $directory = str_replace('vendor/', $this->drupalFinder->getVendorDir(), $directory);
    }

    $uri = $directory . '/' . $file;
    $this->validateWritableComposerFile($uri);

    if (!is_string($data)) {
      $data = json_encode($data, JSON_THROW_ON_ERROR | JSON_PRETTY_PRINT);
    }

    if (file_put_contents($uri, $data) === FALSE) {
      throw new \RuntimeException("The file {$uri} could not be written to.");
    }

    // Reset the static cache.
    $this->cacheBackend->invalidate($uri);
  }

  /**
   * Runs a composer command.
   *
   * Note: USE THIS AT YOUR OWN RISK. YOU ARE RESPONSIBLE FOR ESCAPING THE
   * INPUT TO THE $command PARAMETER.
   *
   * @param string $command
   *   The composer command to run, without the "composer" part.
   * @param string|null $directory
   *   The directory to use as the working directory. If not provided this will
   *   attempt to use DrupalFinder to find the composer root directory.
   * @param mixed $result_code
   *   The result code from running the command, modified by reference.
   *
   * @throws \RuntimeException
   */
  public function exec(string $command, ?string $directory = NULL, mixed &$result_code = NULL): void {
    if (Utility::commandExists('composer')) {
      $directory ??= $this->drupalFinder->getComposerRoot();

      // @todo Should this validate or just let it run and fail?
      $this->validateWritableComposerFile($directory . '/' . static::getComposerFilename());

      // @todo Should this be using passthru() or shell_exec() instead?
      $result = passthru("composer {$command} --working-dir={$directory}", $result_code);
      if ($result === FALSE || $result_code) {
        throw new \RuntimeException("Failed running composer {$command}.");
      }

      // Reset the static cache.
      $this->cacheBackend->invalidateAll();
    }
    else {
      throw new \RuntimeException("The composer command does not exist.");
    }
  }

  /**
   * Validates that the directory is writable or and composer.json is writable.
   *
   * @param string $uri
   *   The URI to the composer.json file.
   *
   * @throws \RuntimeException
   */
  protected function validateWritableComposerFile(string $uri): void {
    if (!is_writable(dirname($uri)) || (is_file($uri) && !is_writable($uri))) {
      throw new \RuntimeException("The file {$uri} does not exist or is not writable.");
    }
  }

}
