<?php

declare(strict_types = 1);

namespace Drupal\Tests\helper\Kernel;

use Drupal\helper\Composer;
use Drupal\KernelTests\KernelTestBase;

/**
 * Tests the Composer helper.
 *
 * @coversDefaultClass \Drupal\helper\Composer
 * @group helper
 */
class ComposerTest extends KernelTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'system',
    'helper',
  ];

  /**
   * Test the composer filename static methods.
   *
   * @covers ::getComposerFilename
   * @covers ::getComposerLockFilename
   */
  public function testComposerFilenames(): void {
    $original = getenv('COMPOSER');
    putenv('COMPOSER');

    $this->assertSame('composer.json', Composer::getComposerFilename());
    $this->assertSame('composer.lock', Composer::getComposerLockFilename());

    putenv('COMPOSER=alternate.ext');
    $this->assertSame('alternate.ext', Composer::getComposerFilename());
    $this->assertSame('alternate.lock', Composer::getComposerLockFilename());

    // Reset the environment back.
    putenv('COMPOSER=' . $original);
  }

  /**
   * Test getting the Drupal composer.json contents.
   *
   * @covers ::getJson
   * @covers ::writeJson
   */
  public function testComposer(): void {
    $composer = new Composer();
    $data = $composer->getJson();
    $this->assertIsArray($data);
    $this->assertArrayHAsKey('require', $data);
    $this->assertNotEmpty($data['require']);

    $vendor_installed = $composer->getJson('vendor/composer/installed.json');
    $this->assertIsArray($vendor_installed);
    $this->arrayHasKey('packages');
    $this->assertNotEmpty($vendor_installed['packages']);

    // Write a new empty composer.json file into the files directory.
    $directory = $this->siteDirectory;
    file_put_contents($directory . '/composer.json', '{}');
    $composer->writeJson([], NULL, $directory);
    $this->assertSame('[]', file_get_contents($directory . '/composer.json'));
    $data = $composer->getJson(NULL, $directory);
    $this->assertSame([], $data);

    $composer->writeJson(['updated' => TRUE], NULL, $directory);
    $this->assertSame("{\n    \"updated\": true\n}", file_get_contents($directory . '/composer.json'));
    $data = $composer->getJson(NULL, $directory);
    $this->assertSame(['updated' => TRUE], $data);
  }

  /**
   * Test reading a file that doesn't exist yet.
   *
   * @covers ::getJson
   */
  public function testFileNotExists() {
    $composer = new Composer();
    $this->expectException(\RuntimeException::class);
    $this->expectExceptionMessage("The file {$this->siteDirectory}/composer.json does not exist or is not readable.");
    $composer->getJson(NULL, $this->siteDirectory);
  }

  /**
   * Test reading an unreadable JSON file.
   *
   * @covers ::getJson
   * @covers ::writeJson
   */
  public function testNotReadable() {
    file_put_contents($this->siteDirectory . '/composer.json', '{}');
    chmod($this->siteDirectory . '/composer.json', 0000);
    $this->expectException(\RuntimeException::class);
    $this->expectExceptionMessage("The file {$this->siteDirectory}/composer.json does not exist or is not readable.");
    $composer = new Composer();
    $composer->getJson(NULL, $this->siteDirectory);
  }

  /**
   * Test reading an invalid JSON file.
   *
   * @covers ::getJson
   * @covers ::writeJson
   */
  public function testInvalidJson(): void {
    $composer = new Composer();
    $composer->writeJson('invalid', NULL, $this->siteDirectory);
    $this->expectException(\JsonException::class);
    $composer->getJson(NULL, $this->siteDirectory);
  }

  /**
   * Returns the actual site directory of the test.
   *
   * Needed for testing exec/passthru commands.
   *
   * @return string
   *   The directory.
   */
  private function getActualDirectory(): string {
    // Use a sub-directory to ensure we don't mess with any other files.
    $directory = $this->root . str_replace('vfs://root', '', $this->siteDirectory) . '/composer-test';
    // Ensure the directory exists and in writable.
    mkdir($directory, 0775, TRUE);
    return $directory;
  }

  /**
   * Test the exec() function.
   *
   * @covers ::exec
   * @covers ::getJson
   */
  public function testExec(): void {
    $directory = $this->getActualDirectory();
    $this->assertDirectoryExists($directory);
    $this->assertDirectoryIsWritable($directory);
    $this->assertFileDoesNotExist($directory . '/composer.json');

    file_put_contents($directory . '/composer.json', '{}');
    $this->assertFileExists($directory . '/composer.json');

    $composer = new Composer();
    $data = $composer->getJson(NULL, $directory);
    $this->assertSame([], $data);

    $composer->exec('config --json extra.test true', $directory, $result_code);
    $this->assertSame(0, $result_code);

    $data = $composer->getJson(NULL, $directory);
    $this->assertSame(['extra' => ['test' => TRUE]], $data);
  }

  /**
   * Test the exec() function without having a composer.json in the directory.
   *
   * @covers ::exec
   */
  public function testExecWithUnwritableDirectory(): void {
    $directory = $this->getActualDirectory();
    chmod($directory, 0000);
    $this->assertDirectoryExists($directory);
    $this->assertDirectoryIsNotWritable($directory);

    $this->expectException(\RuntimeException::class);
    $this->expectExceptionMessage("The file {$directory}/composer.json does not exist or is not writable.");

    $composer = new Composer();
    $composer->exec('config --json extra.test true', $directory, $result_code);
  }

  /**
   * Test the exec() function without having a composer.json in the directory.
   *
   * @covers ::exec
   */
  public function testExecWithUnwritableFile(): void {
    $directory = $this->getActualDirectory();
    file_put_contents($directory . '/composer.json', '{}');
    chmod($directory . '/composer.json', 0000);
    $this->assertDirectoryExists($directory);
    $this->assertDirectoryIsWritable($directory);
    $this->assertFileExists($directory . '/composer.json');
    $this->assertFileIsNotReadable($directory . '/composer.json');

    $this->expectException(\RuntimeException::class);
    $this->expectExceptionMessage("The file {$directory}/composer.json does not exist or is not writable.");

    $composer = new Composer();
    $composer->exec('config --json extra.test true', $directory, $result_code);
  }

}
